import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.clarifai import ClarifaiWriteConfig, SimpleClarifaiConfig


@dataclass
class ClarifaiWriter(Writer):
    write_config: "ClarifaiWriteConfig"
    connector_config: "SimpleClarifaiConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.clarifai import ClarifaiDestinationConnector

        return ClarifaiDestinationConnector
